val version_number : string
val config_file : string ref
val is_native : bool
val logdir : string option ref
val default_user : string ref
val default_group : string ref
val mimefile : string ref
val datadir : string ref
val bindir : string ref
val extdir : string ref
val command_pipe : string ref
val builtin_packages : Ocsigen_lib.String.Set.t
